#include "WProgram.h"
#include "stoplight.h"
#include <avr/io.h>
#include <avr/wdt.h>
#ifdef use_serlcd
#include <SoftwareSerial.h>
#endif

#include <EEPROM.h>
#include "Messenger.h"
#include "Bounce.h"

unsigned long next_tick = 0L;
boolean fire_tick = false;

unsigned long next_powersave_timeout = 0L;

boolean enable_relays = false;

int mode = MODE_SEQUENCE;	// stoplight mode
int state = 0; 				// keep track of where in the sequence 
//int pause = 1;
byte lights = 0;			// what lights are on

int distance_left = 0;
int distance_right = 0;

int distance_left_buff[DISTANCE_BUFF_DEPTH];
int distance_right_buff[DISTANCE_BUFF_DEPTH];

int left_gy;
int left_yg;
int left_yr;
int left_ry;
int right_gy;
int right_yg;
int right_yr;
int right_ry;

unsigned long powersave_timeout = 0L;


boolean button1_state = HIGH;
boolean last_button1_state = HIGH;
boolean button2_state = HIGH;
boolean last_button2_state = HIGH;
boolean button3_state = LOW;
boolean last_button3_state = LOW;
#ifdef BUTTON_4_PIN
boolean button4_state = HIGH;
boolean last_button4_state = HIGH;
#endif

// byte buttons_state = BUTTON_1 | BUTTON_2 | BUTTON_3 | BUTTON_4;
// byte buttons_last_state = BUTTON_1 | BUTTON_2 | BUTTON_3 | BUTTON_4;

#ifdef use_serlcd
SoftwareSerial lcdSerial = SoftwareSerial(LCDRX_PIN, LCDTX_PIN);
#endif

// Instantiate Messenger object with the message function and the 
// default separator (the space character)
Messenger message = Messenger(); 

Bounce button1 = Bounce(BUTTON1_PIN, BUTTON_DEBOUNCE_TIME);
Bounce button2 = Bounce(BUTTON2_PIN, BUTTON_DEBOUNCE_TIME);
Bounce button3 = Bounce(BUTTON3_PIN, MAGSWITCH_DEBOUNCE_TIME);
#ifdef BUTTON_4_PIN
Bounce button4 = Bounce(BUTTON4_PIN, DEBOUNCE_TIME);
#endif

void init_io() {
  // buttons are inputs.
  pinMode(BUTTON1_PIN, INPUT);
  pinMode(BUTTON2_PIN, INPUT);  
  pinMode(BUTTON3_PIN, INPUT);
  //digitalWrite(BUTTON3_PIN, HIGH); // enable internal pullup on button 3
  button3.write(HIGH);               // enable internal pullup on button 3
  button3.update();
  button3.update();
  delay(MAGSWITCH_DEBOUNCE_TIME);
  button3_state = button3.read();
  last_button3_state = button3_state;
  
  
  #ifdef BUTTON_4_PIN
  pinMode(BUTTON4_PIN, INPUT);
  digitalWrite(BUTTON4_PIN, HIGH); // enable internal pullup on button 4
  #endif

  // shift register pins are outputs.
  pinMode(LATCH_PIN, OUTPUT);
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(DATA_PIN, OUTPUT);
  
  update_shift_registers();
}

void init_prefs() {
  enable_relays = EEPROM.read(E2_RELAY_EN_ADDR);     // 0
  mode          = EEPROM.read(E2_DEFAULT_MODE_ADDR); // 0
  left_gy       = EEPROM.read(E2_L_GY_ADDR);         // 25; approx. 2 counts/inch
  left_yg       = left_gy + DISTANCE_HYSTERESIS;
  left_yr       = EEPROM.read(E2_L_YR_ADDR);         // 18
  left_ry       = left_yr + DISTANCE_HYSTERESIS;
  right_gy      = EEPROM.read(E2_R_GY_ADDR);         // 25
  right_yg      = right_gy + DISTANCE_HYSTERESIS;
  right_yr      = EEPROM.read(E2_R_YR_ADDR);         // 18
  right_ry      = right_yr + DISTANCE_HYSTERESIS;
  // powersave_timeout is measured in ms, stored in minutes
  powersave_timeout = 1000L * 60L * EEPROM.read(E2_TIMEOUT_ADDR); // 10
}

void init_distance() {
  // "kick-start" for "Daisy Chaining with Constantly Looping" mode
  // as described in MaxBotix MaxSonar-EZ1 FAQ section 6f:
  // http://www.maxbotix.com/MaxSonar-EZ1__FAQ.html
  pinMode(DISTANCE_GO_PIN, OUTPUT);
  digitalWrite(DISTANCE_GO_PIN, LOW);
  delay(260); // ≥ 250 ms boot time
  digitalWrite(DISTANCE_GO_PIN, HIGH);
  delayMicroseconds(30); // pulse RX pin high for ≥ 20 μs
  // return GO pin to high impedance state
  digitalWrite(DISTANCE_GO_PIN, LOW);
  pinMode(DISTANCE_GO_PIN, INPUT);
  delay(100); // Allow time to calibrate
  // analog pins should now be valid
}

void setup() {
  Serial.begin(9600);
  Serial.println("Starting...");
  
  init_io();
  init_prefs();
  initMessenger();
  init_distance();

  randomSeed(analogRead(DISTANCE_RIGHT_PIN));
  schedule_next_timeout(); // powersave_timeout must be initialized first

  setMode(mode);

  #ifdef use_serlcd
  initLcd();
  lcdSerial.print("Ready");
  Serial.println("Ready");
  #endif
}

// set next powersave timeout
void schedule_next_timeout() {
  next_powersave_timeout = millis() + powersave_timeout;
}

// go into powersave mode now
void timeout_now() {
//  next_powersave_timeout = millis();
  setMode(MODE_POWERSAVE);
}

void reboot() {
  // reset AVR by enabling watchdog and waiting
  // http://support.atmel.no/bin/customer?=&action=viewKbEntry&id=21
  // cli();                  // clear interrupts
  // wdt_enable(WDTO_30MS);  
  // while(1) {};
//    init_prefs();
  setup();
} 

byte relay_translate(byte leds) {
  byte relays = 0;
  if (enable_relays) {
    if (leds & LEFT_RED) {
      relays |= LEFT_RED_RELAY;
    }
    if (leds & LEFT_YELLOW) {
      relays |= LEFT_YELLOW_RELAY;
    }
    if (leds & LEFT_GREEN) {
      relays |= LEFT_GREEN_RELAY;
    }
    if (leds & RIGHT_RED) {
      relays |= RIGHT_RED_RELAY;
    }
    if (leds & RIGHT_YELLOW) {
      relays |= RIGHT_YELLOW_RELAY;
    }
    if (leds & RIGHT_GREEN) {
      relays |= RIGHT_GREEN_RELAY;
    }
  }
  return relays;
}

void update_shift_registers() {
  byte relays = 0;
  relays = relay_translate(lights);
  digitalWrite(LATCH_PIN, LOW);
  shiftOut(DATA_PIN, CLOCK_PIN, MSBFIRST, relays);
  shiftOut(DATA_PIN, CLOCK_PIN, MSBFIRST, lights);
  digitalWrite(LATCH_PIN, HIGH);
}

// schedule how long until the next event
void ticks(int interval) {
  next_tick = millis() + (unsigned long) interval;
}

void nextState() {
  schedule_next_timeout();
  // schedule next event to happen immediately
  ticks(0);
//    Serial.println("state!");
}

void set_relay_enable(int ena) {
  schedule_next_timeout();
  enable_relays = ena;
  Serial.print("Relay enable ");
  Serial.println(enable_relays);
  #ifdef use_serlcd
  lcdClear();
  lcdSerial.print("Relay enable ");
  lcdSerial.print(enable_relays, DEC);
  #endif
}

void toggle_relay_enable() {
  set_relay_enable(!enable_relays);
}

void restore_relay_enable() {
  enable_relays = EEPROM.read(E2_RELAY_EN_ADDR);
}

void handle_buttons() {
  button1.update();
  button2.update();
  button3.update();
  #ifdef BUTTON_4_PIN
  button4.update();
  #endif
  
  button1_state = button1.read();
  button2_state = button2.read();
  button3_state = button3.read();
  #ifdef BUTTON_4_PIN
  button4_state = button4.read();
  #endif
  
  // handle button 1 up: next state
  if (!button1_state && last_button1_state) {
    nextState();
  }

  // handle button 2 up: mode
  else if (!button2_state && last_button2_state) {
    // if button 1 is already down, go into parking mode
    if (!button1_state) { 
	  begin_distance();
	  setMode(MODE_DISTANCE);
	} else {
      setMode(nextMode());
    }
  }

  // handle button 3: go into distance mode
  else if (button3_state && !last_button3_state) {
    begin_distance();
    setMode(MODE_DISTANCE);
  }

  #ifdef BUTTON_4_PIN
  // handle button 4: toggle relay enable
  if (button4_state && !last_button4_state) {
    toggle_relay_enable();
  }
  #endif

  last_button1_state = button1_state;
  last_button2_state = button2_state;
  last_button3_state = button3_state;
  #ifdef BUTTON_4_PIN
  last_button4_state = button4_state;
  #endif
}


void loop() {
  // TODO: handle millisecond rollover every 50 days

  // The following line is the most effective way of 
  // feeding the serial data to Messenger
  while (Serial.available()) {
    byte ch = Serial.read();
    Serial.write(ch); // echo to terminal so we can see what we're typing
    message.process(ch);
  }

  handle_buttons();

  if (millis() > next_powersave_timeout) {
    // mode = MODE_POWERSAVE;
    setMode(MODE_POWERSAVE);
  }

  // is it time to handle the next scheduled event?
  if (millis() > next_tick) {
    fire_tick = true;
  }

  if (fire_tick) {
    switch (mode) {
    case MODE_SEQUENCE:
      handle_sequence();
      break;
    case MODE_RANDOM:
      handle_random();
      break;
    case MODE_SPARKLE:
      handle_sparkle();
      break;
    case MODE_RAILROAD:
      handle_railroad();
      break;
    case MODE_CHASE:
      handle_chase();
      break;
    case MODE_CHASE_REVERSE:
      handle_chase_reverse();
      break;
    case MODE_DISTANCE:
      handle_distance();
      break;
    case MODE_POWERSAVE:
      handle_powersave();
      break;
    } // switch mode

    update_shift_registers();
	
    state += 1;
    fire_tick = false;
  }
}

